/*
 *  renderer - A simple implementation of polygon-based 3D algorithms.
 *  Copyright (C) 2004  Thanassis Tsiodras (ttsiodras@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef _WIN32
#include <windows.h>
#include <GL/glew.h>
#else
#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>
#endif
#include <GL/glut.h>

#include <SDL.h>

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_gl_interop.h>

#include "Types.h"
#include "Base3d.h"
#include "Camera.h"

#include "cudarenderer.h"

__constant__ unsigned VERTICES;
__constant__ unsigned TRIANGLES;

// Utility functions

// CUDA dot product
__device__ coord dotCUDA(const Vector3& l, const Vector3& r)
{
    return l._x*r._x +l._y*r._y +l._z*r._z;
}

// CUDA cross
__device__ Vector3 crossCUDA(const Vector3& l, const Vector3& r)
{
    coord x,y,z;
    const coord &aax=l._x;
    const coord &aay=l._y;
    const coord &aaz=l._z;
    const coord &bbx=r._x;
    const coord &bby=r._y;
    const coord &bbz=r._z;
    x=aay*bbz-bby*aaz;
    y=bbx*aaz-aax*bbz;
    z=aax*bby-aay*bbx;
    return Vector3(x,y,z);
}

// CUDA distance of two points
__device__ coord distanceCUDA(const Vector3& a, const Vector3& b)
{
    coord dx=a._x - b._x;
    coord dy=a._y - b._y;
    coord dz=a._z - b._z;
    return sqrt(dx*dx + dy*dy + dz*dz);
}

// Matrix3x3 multipled by Vector3
__device__ Vector3 multiplyRightWith(const Matrix3& mv, const Vector3& r)
{
    coord xnew = mv._row1._x*r._x + mv._row1._y*r._y + mv._row1._z*r._z;
    coord ynew = mv._row2._x*r._x + mv._row2._y*r._y + mv._row2._z*r._z;
    coord znew = mv._row3._x*r._x + mv._row3._y*r._y + mv._row3._z*r._z;
    return Vector3(xnew, ynew, znew);
}

// Transform Vector3 to any space, given Matrix3 and origin
__device__ Vector3 inline TransformToSomeSpace(Vector3 point, Matrix3 *mv, Vector3 *origin)
{
    point -= *origin;
    return multiplyRightWith(*mv, point);
}

// After transformation in camera space, project and plot (used for point rendering)
#define CLIPPLANEDISTANCE 0.2f

__device__ void inline ProjectAndPlot(const Vector3& xformed, int *pixels, int defaultColor=0x00FFFFFF )
{
    if (xformed._z>CLIPPLANEDISTANCE) {
	int x = (int)(MAXX/2.f  + FOV * xformed._y/xformed._z);
	int y = (int)(MAXY/2.f - FOV * xformed._x/xformed._z);
	if (y>=0.f && y<(int)MAXY && x>=0.f && x<(int)MAXX)
	    pixels[y*MAXX + x] = defaultColor;
    }
}

////////////////////////////////////////
// Rendering function for MODE_POINTS
////////////////////////////////////////

__global__ void CoreLoopVertices(int *pixels, Matrix3 *cudaWorldToCameraSpace, Vertex *pVertices, Vector3 *eye)
{
    unsigned idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx >= VERTICES)
	return;

    // Simple projection and ploting of a white point per vertex

    // Plot projected coordinates (on screen)
    ProjectAndPlot(
	TransformToSomeSpace(pVertices[idx], cudaWorldToCameraSpace, eye),
	pixels);
}

//////////////////////////////////////////////
// Rendering function for MODE_POINTSHIDDEN
//////////////////////////////////////////////

// Create OpenGL BGR value for assignment in PBO buffer
__device__ int getColor(Pixel& p)
{
    return (((unsigned)p._b) << 16) | (((unsigned)p._g) << 8) | (((unsigned)p._r));
}

__global__ void CoreLoopTriangles(int *pixels, Matrix3 *cudaWorldToCameraSpace, Vertex *pVertices, Triangle *pTriangles, Vector3 *eye)
{
    unsigned idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx >= TRIANGLES)
	return;

    // First check if the triangle is visible from where we stand
    // (closed objects only)

    Vector3 triToEye = *eye;
    triToEye -= pTriangles[idx]._center;
    // Normally we would normalize, but since we just need the sign
    // of the dot product (to determine if it facing us or not)...
    //triToEye.normalize();
    if (!pTriangles[idx]._twoSided && dotCUDA(triToEye, pTriangles[idx]._normal)<0.f)
	return;

    // For each of the 3 vertices of triangle j of object i,
    // transform to camera space, project and plot them
    int color = getColor(pTriangles[idx]._colorf);

    ProjectAndPlot(
	TransformToSomeSpace(pVertices[pTriangles[idx]._idx1], cudaWorldToCameraSpace, eye),
	pixels,
	color);
    ProjectAndPlot(
	TransformToSomeSpace(pVertices[pTriangles[idx]._idx2], cudaWorldToCameraSpace, eye),
	pixels,
	color);
    ProjectAndPlot(
	TransformToSomeSpace(pVertices[pTriangles[idx]._idx3], cudaWorldToCameraSpace, eye),
	pixels,
	color);
}

///////////////////////////////////////////////
// MODE_RAYCASTER and MODE_RAYCASTER_SHADOWS
///////////////////////////////////////////////

// The triangle specific light-space bounding box, plus a "pointer" to the triangle
// (index in the triangle array)
struct BoundingBoxLight {
    coord _bx,_by,_tx,_ty;
    int _triIdx;
};

// The triangle specific camera-space bounding box, plus a "pointer" to the triangle
// (index in the triangle array). It also carries the camera-space coordinates of
// the 3 vertices, as well as the camera-space normals. These are pre-calculated
// during the first CUDA kernel (CalculateTrianglesCameraBoundingBoxes) and used
// during the final one (CoreLoopTrianglesRaycaster)
struct BoundingBoxCamera {
    coord _bx,_by,_tx,_ty;
    int _triIdx;
    Vector3 _xformedA, _xformedB, _xformedC;
    Vector3 _xformedNrmA, _xformedNrmB, _xformedNrmC;
};

// pre-calculated information is created before CoreLoopTrianglesRaycaster is called.
// For the lightspace, the per-triangle information kept is the one used for ray/triangle intersections
struct IntersectionCacheLightspace {
    Vector3 _nrm, _e1, _e2, _e3;
    coord _d, _d1, _d2, _d3;
};

// For the cameraspace, the per-triangle information kept is the one used for ray/triangle intersections
// plus a pointer to the the triangle's bounding box. Why? Because the bounding box includes pre-calculated
// vertex and vertex normals in camera space.
struct IntersectionCacheCameraspace {
    Vector3 _nrm, _e1, _e2, _e3;
    coord _d, _d1, _d2, _d3;
    unsigned _bbIdx;
};

// 1st CUDA kernel:
// 
// Does two jobs: 
//  - projects the triangles and calculates their bounding boxes, 
//  - and stores the projected vertices/normals as well as the pre-calculated ray/triangle intersection data
__global__ void CalculateTrianglesCameraBoundingBoxes(
    Matrix3 *cudaWorldToCameraSpace,
    Vertex *pVertices, Triangle *pTriangles,
    Vector3 *cudaEyePosInWorldSpace,
    int *ptrCudaCameraBoundingBoxesCount,
    BoundingBoxCamera* ptrCudaCameraBBoxes, IntersectionCacheCameraspace *ptrCudaCameraIntersectionsCachePerTriangle)
{
    unsigned idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx >= TRIANGLES)
	return;

    // Is this triangle visible?

    Vector3 triToEye = *cudaEyePosInWorldSpace;
    triToEye -= pTriangles[idx]._center;
    // Normally we would normalize, but since we just need the sign
    // of the dot product (to determine if it facing us or not)...
    //triToEye.normalize();
    if (!pTriangles[idx]._twoSided && dotCUDA(triToEye, pTriangles[idx]._normal)<0.f)
	return;

    // Transform vertices to camera space
    Vertex& p1 = pVertices[pTriangles[idx]._idx1];
    Vector3 xformed_a = TransformToSomeSpace(p1, cudaWorldToCameraSpace, cudaEyePosInWorldSpace);
    if (xformed_a._z < CLIPPLANEDISTANCE) return;

    Vertex& p2 = pVertices[pTriangles[idx]._idx2];
    Vector3 xformed_b = TransformToSomeSpace(p2, cudaWorldToCameraSpace, cudaEyePosInWorldSpace);
    if (xformed_b._z < CLIPPLANEDISTANCE) return;

    Vertex& p3 = pVertices[pTriangles[idx]._idx3];
    Vector3 xformed_c = TransformToSomeSpace(p3, cudaWorldToCameraSpace, cudaEyePosInWorldSpace);
    if (xformed_c._z < CLIPPLANEDISTANCE) return;

    // project to 2D screen

    coord ax,ay, bx,by, cx,cy;

#define PROJECTY(p) \
    p##y = MAXY/2.f - FOV * xformed_##p ._x / xformed_##p ._z;

    PROJECTY(a)
    PROJECTY(b)
    PROJECTY(c)

    if (ay<0 && by<0 && cy<0) return;
    if (ay>=MAXY && by>=MAXY && cy>=MAXY) return;

    // The triangle is (at least partially) visible; it will be therefore added to the bounding box list

#define PROJECTX(p) \
    p##x = MAXX/2.f  + FOV * xformed_##p ._y / xformed_##p ._z;

    PROJECTX(a)
    PROJECTX(b)
    PROJECTX(c)

    // ...and reset the triangle's bounding box (in projected screen space).
    coord minx,miny,maxx,maxy;
    minx=miny=1e10;
    maxx=maxy=-1e10;

    // This macro updates this triangle's bounding box
#define BBOX(p)						\
    minx = fminf(minx,p ## x);				\
    miny = fminf(miny,p ## y);				\
    maxx = fmaxf(maxx,p ## x);				\
    maxy = fmaxf(maxy,p ## y);

    BBOX(a)
    BBOX(b)
    BBOX(c)

    // And store it to an atomically chosen new index in the bounding boxes array
    //
    // Why are we doing this atomic add and not just use [idx] ?
    // We are doing this per triangle, so we would - if ALL the triangles were visible.
    // But since some of them are filtered out because of backface culling, we only "reserve" 
    // (and later, work on) as many as are visible.
    int bbIdx = atomicAdd(ptrCudaCameraBoundingBoxesCount, 1);
    BoundingBoxCamera& box = ptrCudaCameraBBoxes[bbIdx];

    // Store the bounding box itself
    box._bx = minx;
    box._by = miny;
    box._tx = maxx;
    box._ty = maxy;

    // Store a "pointer" to the triangle that generated this bbox
    box._triIdx = idx;

    // Store the transformed (to camera space) vertices of the triangle
    box._xformedA = xformed_a;
    box._xformedB = xformed_b;
    box._xformedC = xformed_c;

    // Store the transformed (to camera space) vertice normals
    box._xformedNrmA = multiplyRightWith(*cudaWorldToCameraSpace, p1._normal);
    box._xformedNrmB = multiplyRightWith(*cudaWorldToCameraSpace, p2._normal);
    box._xformedNrmC = multiplyRightWith(*cudaWorldToCameraSpace, p3._normal);

    // Now pre-calculate triangle-specific intersection-logic data...
    // Algorithm for triangle intersection is taken from Roman Kuchkuda's paper.

    Vector3 vc1=xformed_b; vc1-=xformed_a; // edge vectors
    Vector3 vc2=xformed_c; vc2-=xformed_b;
    Vector3 vc3=xformed_a; vc3-=xformed_c;

    // plane of triangle
    Vector3 nrm = crossCUDA(vc1, vc2); nrm.normalizeCUDA();
    coord d = dotCUDA(nrm, xformed_a);

    // edge planes
    Vector3 e1 = crossCUDA(nrm, vc1); e1.normalizeCUDA();
    coord d1 = dotCUDA(e1, xformed_a);
    Vector3 e2 = crossCUDA(nrm, vc2); e2.normalizeCUDA();
    coord d2 = dotCUDA(e2, xformed_b);
    Vector3 e3 = crossCUDA(nrm, vc3); e3.normalizeCUDA();
    coord d3 = dotCUDA(e3, xformed_c);

    // And store the precalculated camera-space intersection data per triangle in the "cache" array

    #define CACHE(x) \
    ptrCudaCameraIntersectionsCachePerTriangle[idx]._ ## x = x;
    CACHE(nrm)
    CACHE(e1)
    CACHE(e2)
    CACHE(e3)
    CACHE(d)
    CACHE(d1)
    CACHE(d2)
    CACHE(d3)
    CACHE(bbIdx)
}

// Same as above, but for light space.
// (stores less pre-computed info in the per-triangle ray-intersection data)
__global__ void CalculateTrianglesLightBoundingBoxes(
    Matrix3 *cudaWorldToLightSpace,
    Vertex *pVertices, Triangle *pTriangles,
    Vector3 *cudaLightPosInWorldSpace,
    int *ptrCudaLightBoundingBoxesCount,
    BoundingBoxLight* ptrCudaLightBBoxes, IntersectionCacheLightspace *ptrCudaLightIntersectionsCachePerTriangle)
{
    unsigned idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx >= TRIANGLES)
	return;

    // backface culling for light-space
    // Is this triangle visible?

    Vector3 triToEye = *cudaLightPosInWorldSpace;
    triToEye -= pTriangles[idx]._center;
    // Normally we would normalize, but since we just need the sign
    // of the dot product (to determine if it facing us or not)...
    //triToEye.normalize();
    if (!pTriangles[idx]._twoSided && dotCUDA(triToEye, pTriangles[idx]._normal)<0.f)
	return;

    // Transform vertices to light space
    Vertex& p1 = pVertices[pTriangles[idx]._idx1];
    Vector3 xformed_a = TransformToSomeSpace(p1, cudaWorldToLightSpace, cudaLightPosInWorldSpace);
    if (xformed_a._z < CLIPPLANEDISTANCE) return;

    Vertex& p2 = pVertices[pTriangles[idx]._idx2];
    Vector3 xformed_b = TransformToSomeSpace(p2, cudaWorldToLightSpace, cudaLightPosInWorldSpace);
    if (xformed_b._z < CLIPPLANEDISTANCE) return;

    Vertex& p3 = pVertices[pTriangles[idx]._idx3];
    Vector3 xformed_c = TransformToSomeSpace(p3, cudaWorldToLightSpace, cudaLightPosInWorldSpace);
    if (xformed_c._z < CLIPPLANEDISTANCE) return;

    // project to 2D "light-screen" (think of where a shadowmap "screen" would live)

    coord ax,ay, bx,by, cx,cy;

    // In the projection of light, I use a scaling factor of 1024.
    // This is of course arbitrary - the point is that the resulting projected coordinates
    // will be used to calculate the "perspective projected light space bounding box",
    // and since we are using floats and not doubles (not yet in CUDA) we need every bit
    // of accuracy we can get. If the LightDistanceFactor in main.cpp is increased,
    // then this 1024 must also be increased, to always make sure we have enough "space"
    // for our numbers to be placed within the accuracy of a "float".

#undef PROJECTY
#define PROJECTY(p) \
    p##y = 1024.f * xformed_##p ._x / xformed_##p ._z;

    PROJECTY(a)
    PROJECTY(b)
    PROJECTY(c)

    // The triangle  will be added now to the bounding box list
#undef PROJECTX
#define PROJECTX(p) \
    p##x = 1024.f * xformed_##p ._y / xformed_##p ._z;

    PROJECTX(a)
    PROJECTX(b)
    PROJECTX(c)

    // reset the triangle's bounding box (in projected screen space).
    coord minx,miny,maxx,maxy;
    minx=miny=1e10;
    maxx=maxy=-1e10;

    // This macro updates this triangle's bounding box
#undef BBOX
#define BBOX(p)						\
    minx = fminf(minx,p ## x);				\
    miny = fminf(miny,p ## y);				\
    maxx = fmaxf(maxx,p ## x);				\
    maxy = fmaxf(maxy,p ## y);

    BBOX(a)
    BBOX(b)
    BBOX(c)

    // And store the bounding box to an atomically chosen new index in the bounding boxes array
    //
    // Why are we doing this atomic add and not just use [idx] ?
    // We are doing this per triangle, so we would - if ALL the triangles were visible.
    // But since some of them are filtered out because of backface culling, we only "reserve" 
    // (and later, work on) as many as are visible.
    int bbIdx = atomicAdd(ptrCudaLightBoundingBoxesCount, 1);
    BoundingBoxLight& box = ptrCudaLightBBoxes[bbIdx];

    // Store the bounding box coordinates
    box._bx = minx;
    box._by = miny;
    box._tx = maxx;
    box._ty = maxy;

    // Store a "pointer" to the triangle that generated this bbox
    box._triIdx = idx;

    // Now pre-calculate triangle-specific intersection-logic data (in light space!)...
    // Algorithm for triangle intersection is taken from Roman Kuchkuda's paper.

    Vector3 vc1=xformed_b; vc1-=xformed_a; // edge vectors
    Vector3 vc2=xformed_c; vc2-=xformed_b;
    Vector3 vc3=xformed_a; vc3-=xformed_c;

    // plane of triangle
    Vector3 nrm = crossCUDA(vc1, vc2); nrm.normalizeCUDA();
    coord d = dotCUDA(nrm, xformed_a);

    // edge planes
    Vector3 e1 = crossCUDA(nrm, vc1); e1.normalizeCUDA();
    coord d1 = dotCUDA(e1, xformed_a);
    Vector3 e2 = crossCUDA(nrm, vc2); e2.normalizeCUDA();
    coord d2 = dotCUDA(e2, xformed_b);
    Vector3 e3 = crossCUDA(nrm, vc3); e3.normalizeCUDA();
    coord d3 = dotCUDA(e3, xformed_c);

    // And store the precalculated light-space intersection data per triangle in the "cache" array

#undef CACHE
#define CACHE(x) \
    ptrCudaLightIntersectionsCachePerTriangle[idx]._ ## x = x;

    CACHE(nrm)
    CACHE(e1)
    CACHE(e2)
    CACHE(e3)
    CACHE(d)
    CACHE(d1)
    CACHE(d2)
    CACHE(d3)
}

// 2nd CUDA kernel:
// 
// The 1st kernel performed the transformations of vertices and normals to the proper spaces,
// and thus, computed the bounding box of each triangle (in the 2D screen and "light-screen" space).
// We must now compute the complete scene's bounding box:
// 
// Parallel reduction from N bounding boxes... down to one.
// (N is how many triangles were visible in this particular frame)

template <class BoundingBox>
__global__ void ParallelReductionOfBoundingBoxes(
    BoundingBox *ptrCudaBBoxes, BoundingBox *ptrCudaBBoxesPerThreadBlock, int *ptrCudaBoundingBoxesCount)
{
    // Use the fast shared mem to do the comparisons
    __shared__ BoundingBox sdata[THREADS_PER_BLOCK];

    // load shared mem
    unsigned int tid = threadIdx.x;
    unsigned int i = blockIdx.x*blockDim.x + threadIdx.x;

    if (i<*ptrCudaBoundingBoxesCount)
	sdata[tid] = ptrCudaBBoxes[i];
    else {
	// for the last block of threads, fill up with impossible bounding boxes, so they are not selected
	sdata[tid]._bx = 1e10;
	sdata[tid]._by = 1e10;
	sdata[tid]._tx = -1e10;
	sdata[tid]._ty = -1e10;
    }
    __syncthreads();

    // do reduction in shared mem
    for(unsigned int s=blockDim.x/2; s>0; s>>=1) {
        if (tid < s) {
            sdata[tid]._bx = fminf(sdata[tid]._bx, sdata[tid + s]._bx);
            sdata[tid]._by = fminf(sdata[tid]._by, sdata[tid + s]._by);
            sdata[tid]._tx = fmaxf(sdata[tid]._tx, sdata[tid + s]._tx);
            sdata[tid]._ty = fmaxf(sdata[tid]._ty, sdata[tid + s]._ty);
        }
        __syncthreads();
    }

    // Have thread 0 in the block write the result for this block's bounding box (to global mem)
    if (tid == 0) ptrCudaBBoxesPerThreadBlock[blockIdx.x] = sdata[0];
}

// 3rd CUDA kernel:
// 
// The 2nd kernel found the scene's total bounding box.
// We can now see which cell (in the RAYCASTBINS x RAYCASTBINS array) each bounding box falls in.
//
// But we can't place it - not yet.
// We just measure how many bounding boxes will be allocated in each cell:

template <class BoundingBox>
__global__ void CountBinTriangleListDepths(
    coord binSpanX, coord binSpanY, coord bx, coord by,
    BoundingBox *ptrCudaBBoxes, unsigned *ptrCudaBinsGridCounts, int *ptrCudaBoundingBoxesCount)
{
    //__shared__ int bins[RAYCASTBINS*RAYCASTBINS];
    unsigned idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx >= *ptrCudaBoundingBoxesCount)
	return;

    // Find which "cell" in the RAYCASTBINS x RAYCASTBINS grid this bbox falls in.
    int bottomx = int((ptrCudaBBoxes[idx]._bx - bx)/binSpanX);
    int bottomy = int((ptrCudaBBoxes[idx]._by - by)/binSpanY);
    int topx = int((ptrCudaBBoxes[idx]._tx - bx)/binSpanX);
    int topy = int((ptrCudaBBoxes[idx]._ty - by)/binSpanY);

    // Make sure it is within bounds (it should, but floating point slight deviations may cause harm)
#define CLAMP(idx)	    \
    if (idx<0)		    \
	idx=0;		    \
    if (idx>=RAYCASTBINS)   \
	idx=RAYCASTBINS-1;

    CLAMP(bottomx);
    CLAMP(bottomy);
    CLAMP(topx);
    CLAMP(topy);

    // For all the bins that the triangle "spans" across, atomically increment a counter
    for(int a=bottomy; a<=topy; a++)
	for(int b=bottomx; b<=topx; b++)
	    atomicAdd(&ptrCudaBinsGridCounts[a*RAYCASTBINS+b], 1);
}

// 4th CUDA kernel:
// 
// The 3rd kernel found how many bboxes will be placed in each of the RAYCASTBINS x RAYCASTBINS cells.
// We will now accumulate them, so that...
//
// 0 1 0 3
// 1 2 0 4
// ....
//
// will become:
//
// 0 1 1 4
// 5 7 7 11
//
// i.e. we will find out two things: (a) the total cells we will need (the number in the bottom right cell.
// after this accumulation), and (b) the final "offset" of the last "entry" in each grid cell.
// 
// This could be optimized a lot, but the profiler clearly shows it is of no consequence.
//
__global__ void CalculateBinOffsets(
    unsigned* ptrCudaBinsGridCounts, unsigned *ptrCudaBinsGridOffsets, int *ptrCudaLastBinOffset)
{
    unsigned idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx >= RAYCASTBINS*RAYCASTBINS)
	return;

    int total = 0;
    for(int i=0; i<=idx; i++)
	total += ptrCudaBinsGridCounts[i];

    ptrCudaBinsGridOffsets[idx] = total;

    if (idx == RAYCASTBINS*RAYCASTBINS - 1)
	*ptrCudaLastBinOffset = total;
}

// 5th CUDA kernel:
//
// Finally, we've allocated as many cells as we were told in the bottom right corner of ptrCudaBinsGridOffsets.
// And we can now place each bounding box in the bin's "list". The ptrCudaBinsGridOffsets tells us where to
// store it.
// 
template <class BoundingBox>
__global__ void PlaceTrianglesInBins(
    coord binSpanX, coord binSpanY, coord bx, coord by,
    BoundingBox *ptrCudaBBoxes, unsigned *ptrCudaBinsGridOffsets, int *ptrCudaBins, int *ptrCudaBoundingBoxesCount)
{
    unsigned int idx = blockIdx.x*blockDim.x + threadIdx.x;

    if (idx>=*ptrCudaBoundingBoxesCount)
	return;

    int bottomx = int((ptrCudaBBoxes[idx]._bx - bx)/binSpanX);
    int bottomy = int((ptrCudaBBoxes[idx]._by - by)/binSpanY);
    int topx = int((ptrCudaBBoxes[idx]._tx - bx)/binSpanX);
    int topy = int((ptrCudaBBoxes[idx]._ty - by)/binSpanY);
    CLAMP(bottomx);
    CLAMP(bottomy);
    CLAMP(topx);
    CLAMP(topy);
    // For all the bins that the triangle "spans" across, store the triangle's info in the proper bin list.
    int triIdx = ptrCudaBBoxes[idx]._triIdx;
    for(int a=bottomy; a<=topy; a++)
	for(int b=bottomx; b<=topx; b++)
	{
	    int i = (int) atomicSub(&ptrCudaBinsGridOffsets[a*RAYCASTBINS+b],1) - 1;
	    ptrCudaBins[i] = triIdx; // There - store the triangle index, which is all we need anyway.
	}
}

// 5th CUDA kernel:
//
// (the beast)
//
// Templated, to work for both plain Phong illumination and PhongAndShadows (doShadows is compile-time param)
//
template <bool doShadows>
__global__ void CoreLoopTrianglesRaycaster(
    int *pixels, Matrix3 *cudaWorldToCameraSpace, Matrix3 *cudaCameraToLightSpace,
    Vertex *pVertices, Triangle *pTriangles,
    Vector3 *cudaEyePosInWorldSpace, Vector3 *cudaLightPosInWorldSpace, Vector3 *cudaLightPosInCameraSpace,
    coord binCameraSpanX, coord binCameraSpanY, coord bCamerax, coord bCameray,
    coord binLightSpanX, coord binLightSpanY, coord bLightx, coord bLighty,
    int *ptrCudaCameraBins, unsigned *ptrCudaCameraBinsGridOffsets, int *ptrCudaCameraBoundingBoxesCount,
    int *ptrCudaLightBins, unsigned *ptrCudaLightBinsGridOffsets, int *ptrCudaLightBoundingBoxesCount,
    IntersectionCacheCameraspace *ptrCudaCameraIntersectionsCachePerTriangle, BoundingBoxCamera *ptrCudaCameraBBoxes,
    IntersectionCacheLightspace *ptrCudaLightIntersectionsCachePerTriangle, BoundingBoxLight *ptrCudaLightBBoxes,
    int *ptrCudaCameraLastBinOffset, int *ptrCudaLightLastBinOffset)
{
    // To be more coherent (cache-wise), this code will not allocate idx to a normal, left-to-right, top-to-bottom 
    // allocation of pixels. Instead, it will navigate in screen space in a way that fills up the first bin first,
    // then the 2nd, etc. This way, the threads are reading from the same triangle data, which *should* help speed.

    // Only one thread in the block will write data to the global buffer - but since we can't say "threadIdx.x == 0"
    // (because thread 0 may simply return quickly, if it intercepts no triangle) we use an atomically accessed flag
    // - the first thread that reaches it, writes the data to the pixels buffer, in a nicely coalesced manner
    // (look at the end of this kernel)
    int __shared__ flagWriteResults;

    // The threads will store their results in shared memory, and the flagWriteResults will be used to trigger
    // one thread to copy the pixel data to the global frame buffer (pixels) at the end of this kernel.
    unsigned __shared__ binPixels[RAYCASTBINPIXELS];

    unsigned int idx = blockIdx.x*blockDim.x + threadIdx.x;
    if (idx>=MAXX*MAXY)
	return;

    flagWriteResults = 0;

    // Start with AMBIENT (16.f)
    binPixels[threadIdx.x] = 0x404040;

    // try to be better cache-coherent
    // 
    // RAYCASTBINPIXELS = RAYCASTBINSW * RAYCASTBINSH
    //                  = MAXX/RAYCASTBINS * MAXY/RAYCASTBINS

    int ii = idx % RAYCASTBINPIXELS; // index inside the bin, 0 .. RAYCASTBINPIXELS - 1
    int jj = idx / RAYCASTBINPIXELS; // index of bin, from 0 .. RAYCASTBINS^2 - 1 

    // remember, we have RAYCASTBINS x RAYCASTBINS bins per screen
    int x = (jj % RAYCASTBINS) * RAYCASTBINSW + ( ii % RAYCASTBINSW );
    int y = (jj / RAYCASTBINS) * RAYCASTBINSH + ( ii / RAYCASTBINSW );

    // In theory, this should have helped... but it didn't.
    //if (x<int(bCamerax) || y<int(bCameray) || x>int(tCamerax+1) || y>int(tCameray+1))
    //	return;

    // the ray from the eye to the screen's (x,y), has this direction:
    coord lx = coord((MAXY/2)-y)/FOV;
    coord ly = coord(x-(MAXX/2))/FOV;
    coord lz = 1.0f;
    Vector3 ray = Vector3(lx,ly,lz);
    ray.normalizeCUDA();

    // use the bin span (calculated via the 2nd kernel) to see which bin we "fall" on:
    int binx = (int)((x - bCamerax)/binCameraSpanX);
    int biny = (int)((y - bCameray)/binCameraSpanY);
    CLAMP(binx)
    CLAMP(biny)

    // The list of triangles in this bin, is accessed via the offsets array:
    // offset[biny][binx] contains the index of the first triangle in the list.
    unsigned startingOffset = ptrCudaCameraBinsGridOffsets[biny*RAYCASTBINS + binx];
    // The neighbouring cell contains the upper-bound
    unsigned endingOffset;
    if ((biny == RAYCASTBINS-1) && (binx == RAYCASTBINS-1))
	endingOffset = *ptrCudaCameraLastBinOffset;
    else
	endingOffset = ptrCudaCameraBinsGridOffsets[biny*RAYCASTBINS + binx + 1];

    int bestTriIdx = -1;
    coord bestTriDist = 1e10;
    Vector3 pointHitInCameraSpace;
    coord kAB=0.f, kBC=0.f, kCA=0.f; // distances from the 3 edges of the point where we hit

    // For all the triangles in the bin of this pixel, check whether the ray pierces the triangle
    for (int i=startingOffset; i<endingOffset; i++) {
	int triIdx = ptrCudaCameraBins[i];

	// use the pre-calculated intersection data (that the 1st kernel prepared for us)
#define CACHED(x) (ptrCudaCameraIntersectionsCachePerTriangle[triIdx]._ ## x)
	
	coord k = dotCUDA(CACHED(nrm), ray);
	if (k == 0.0f)
	    continue; // this triangle is parallel to the ray - can't be intersected.

	//coord s = (d - dot(nrm,origin))/k; // but since origin is (0,0,0)...
	coord s = CACHED(d)/k;
	if (s <= 0.0f) // this triangle is behind us.
	    continue;

	Vector3 hit = ray;
	hit *= s;

	// Is the intersection of the ray with the triangle's plane INSIDE the triangle?
	coord kt1 = dotCUDA(CACHED(e1), hit) - CACHED(d1); if (kt1<0.0f) continue;
	coord kt2 = dotCUDA(CACHED(e2), hit) - CACHED(d2); if (kt2<0.0f) continue;
	coord kt3 = dotCUDA(CACHED(e3), hit) - CACHED(d3); if (kt3<0.0f) continue;

	// It is, "hit" is the cameraspace coordinate of the intersection.
	// check it's Z coordinate to see if it is the closest one we've seen...
	if (hit._z < bestTriDist) {
	    bestTriIdx = triIdx;
	    bestTriDist = hit._z;
	    pointHitInCameraSpace = hit;
	    kAB = kt1; // keep the distance from the edges, we will use them in Phong interpolation
	    kBC = kt2;
	    kCA = kt3;
	}
    }

    if (bestTriIdx == -1) {
	return; // no triangle hit, so: ambient (already set above, binPixels[threadIdx.x] = 0x404040;
    }

    // This is the closest triangle our ray hit:

    Triangle *pBestTri = &pTriangles[bestTriIdx];

    // Instead of using the pre-calculated camera space triangle normal
    // perform Phong interpolation...

    BoundingBoxCamera& box = ptrCudaCameraBBoxes[ptrCudaCameraIntersectionsCachePerTriangle[bestTriIdx]._bbIdx];

    // These are the triangle's vertices camera space coordinates:
    Vector3& xformedA = box._xformedA;
    Vector3& xformedB = box._xformedB;
    Vector3& xformedC = box._xformedC;

    // And these are the triangle's vertices camera space normals:
    Vector3& xformedNrmA = box._xformedNrmA;
    Vector3& xformedNrmB = box._xformedNrmB;
    Vector3& xformedNrmC = box._xformedNrmC;

    // We now want to interpolate the triangle's normal,
    // so that as the "pointHitInCameraSpace" gets closer to
    // a vertex X, the interpolated normal becomes closer to xformedNrmX,
    // and becomes EXACTLY xformedNrmX, if the pointHitInCameraSpace is X.
    //
    // To do that, we use the 3 areas of the triangle, as it is divided
    // by the pointHitInCameraSpace.
    //
    // This is the total triangle's area: cross product of two edges
    // (in fact, we should divide by 2, but since we're only interested
    //  in ratios (see below), there is no need)
    Vector3 AB = xformedB; AB-= xformedA;  // edge AB in camera space
    Vector3 BC = xformedC; BC-= xformedB;  // edge BC in camera space
    Vector3 crossAB_BC = crossCUDA(AB, BC);
    coord area = crossAB_BC.length();      // 2*area(ABC)

    // And these are the three sub-triangles - kAB,kBC,kCA were found above...
    coord ABx = kAB*distanceCUDA(xformedA, xformedB); // again, these are 2*area...
    coord BCx = kBC*distanceCUDA(xformedB, xformedC);
    coord CAx = kCA*distanceCUDA(xformedC, xformedA);

    // use the area of the sub-triangle ACROSS a point, to scale!
    // (which means that if pointHitInCameraSpace is exactly on a vertex,
    //  the area of the sub-triangle becomes the area of the triangle!)
    Vector3 phongNormalA = xformedNrmA; phongNormalA *= BCx / area;
    Vector3 phongNormalB = xformedNrmB; phongNormalB *= CAx / area;
    Vector3 phongNormalC = xformedNrmC; phongNormalC *= ABx / area;

    // and finally, accumulate the three contributions and normalize.
    Vector3 phongNormal = phongNormalA;
    phongNormal += phongNormalB;
    phongNormal += phongNormalC;
    phongNormal.normalizeCUDA();

    // OK, we hit a triangle - but are we in shadow?
    // Let's also compute the final color
    Pixel target = pBestTri->_colorf; // start from the triangle color in ambient:
    Vertex& p1 = pVertices[pBestTri->_idx1];
    Vertex& p2 = pVertices[pBestTri->_idx2];
    Vertex& p3 = pVertices[pBestTri->_idx3];
    coord ambientOcclusionCoeff = 
	p1._ambientOcclusionCoeff * BCx/area +
	p2._ambientOcclusionCoeff * CAx/area +
	p3._ambientOcclusionCoeff * ABx/area;
    coord ambient = ((AMBIENT*ambientOcclusionCoeff/255.0f)/255.0f);
    target *= ambient;

    // This light's diffuse and specular contribution
    Pixel dColor = Pixel(); // start with black

    Vector3 pointToLightInCameraSpace = *cudaLightPosInCameraSpace;
    pointToLightInCameraSpace -= pointHitInCameraSpace;

    // we transform the point hit to light space coordinates (inLightSpace)
    Vector3 lightToPointInCameraSpace = pointHitInCameraSpace;
    lightToPointInCameraSpace -= *cudaLightPosInCameraSpace;
    Vector3 inLightSpace = multiplyRightWith(*cudaCameraToLightSpace, lightToPointInCameraSpace);

    if (inLightSpace._z <0.f) return; // we hit a place which is behind our light? ambient...

    // We need the vector from light to point:
    Vector3 pointToLight = inLightSpace;
    pointToLight *= -1.f;

    bool inShadow = false;
    if (doShadows) { // compile-time param from the template

	// for all the triangles in the scene *that our "inLightSpace" point MAY be obstructed by*,
	//   see if the ray cast from inLightSpace to Light pierces the triangle
	//   if it does, check whether the Z of the intersection is less than inLightSpace.Z
	//   if it is, inShadow = True and break out of the loop

	// To see which bins contain the triangles we MAY be obstructed by,
	// we project inLightSpace, and use the pre-calculated bins...

	// Again, in the projection of light, I use a scaling factor of 1024.
	// This is of course arbitrary - the point is that the resulting projected coordinates
	// will be used to calculate the "perspective projected light space bounding box",
	// and since we are using floats and not doubles (not yet in CUDA) we need every bit
	// of accuracy we can get. If the LightDistanceFactor in main.cpp is increased,
	// then this 1024 must also be increased, to always make sure we have enough "space"
	// for our numbers to be placed within the accuracy of a "float".

	lx = 1024.f*inLightSpace._y/inLightSpace._z;
	ly = 1024.f*inLightSpace._x/inLightSpace._z;

	// Find the bin this corresponds to
	binx = (int)((lx - bLightx)/binLightSpanX);
	biny = (int)((ly - bLighty)/binLightSpanY);
	CLAMP(binx)
	CLAMP(biny)

	// instead of checking ALL triangles to see if they block our light, we only check those 
	// that MAY block us.
	//
	// The list of triangles in this bin, is accessed via the offsets array:
	// offset[biny][binx] contains the index of the first triangle in the list.
	startingOffset = ptrCudaLightBinsGridOffsets[biny*RAYCASTBINS + binx];
	if ((biny == RAYCASTBINS-1) && (binx == RAYCASTBINS-1))
	    endingOffset = *ptrCudaLightLastBinOffset;
	else
	    // The neighbouring cell contains the upper-bound
	    endingOffset = ptrCudaLightBinsGridOffsets[biny*RAYCASTBINS + binx + 1];

	ray = pointToLight;
	ray.normalizeCUDA();

	// For all the triangles in the bin of this "light" pixel, check whether the ray pierces the triangle
	for (int i=startingOffset; i<endingOffset; i++) {
	    int triIdx = ptrCudaLightBins[i];

    #undef CACHED
    #define CACHED(x) (ptrCudaLightIntersectionsCachePerTriangle[triIdx]._ ## x)

	    // ray from inLightSpace to light...
	    coord k = dotCUDA(CACHED(nrm), ray);
	    if (k == 0.0f)
		continue; // this triangle is parallel to the light ray - can't cast shadow.

	    coord s = (CACHED(d) - dotCUDA(CACHED(nrm),inLightSpace))/k;
	    if (s <= 0.0f) // this triangle is "behind" the "inLightSpace" point. Can't cast shadow.
		continue;

	    Vector3 hit = ray;
	    hit *= s;
	    hit += inLightSpace;

	    // Is the intersection of the ray with the triangle's plane INSIDE the triangle?
	    k = dotCUDA(CACHED(e1), hit) - CACHED(d1); if (k<0.0f) continue;
	    k = dotCUDA(CACHED(e2), hit) - CACHED(d2); if (k<0.0f) continue;
	    k = dotCUDA(CACHED(e3), hit) - CACHED(d3); if (k<0.0f) continue;

	    // It is, "hit" is the lightspace coordinate of the intersection.
	    // check it's Z coordinate... and use a bit of bias to avoid self shadowing
	    if (hit._z < inLightSpace._z - 0.00001f) {
		inShadow = true;
		break;
	    }
	}
    }

    // time to do Phong lighting

    // Diffuse color
    pointToLightInCameraSpace.normalizeCUDA();  // vector from point to light (in camera space)
    coord intensity = dotCUDA(phongNormal, pointToLightInCameraSpace);
    if (inShadow || (intensity<0.f)) {
	; // in shadow, let it be in ambient
    } else {
	Pixel diffuse = pBestTri->_colorf;
	diffuse *= (coord) (DIFFUSE*intensity/255.f);   // diffuse contribution
	dColor += diffuse;

	// Specular color
	// We will use the half vector: pointToLightInCameraSpace + point to camera
	// Since 'point' now has camera space coordinates, -1.0*point is in fact,
	// a vector from point to camera.
	Vector3 pointToCamera = pointHitInCameraSpace;
	pointToCamera *= -1.f;
	pointToCamera.normalizeCUDA();

	Vector3 half = pointToLightInCameraSpace;
	half += pointToCamera;
	half.normalizeCUDA();

	coord intensity2 = dotCUDA(half, phongNormal);
	if (intensity2>0.f) {
	    intensity2 *= intensity2;
	    intensity2 *= intensity2;
	    intensity2 *= intensity2;
	    intensity2 *= intensity2;
	    intensity2 *= intensity2;
	    dColor += Pixel(
		(unsigned char)(SPECULAR*intensity2),
		(unsigned char)(SPECULAR*intensity2),
		(unsigned char)(SPECULAR*intensity2));
	}
	target += dColor; // specular contribution
    }

    // Clamp result to proper range
    if (target._r>255.0f) target._r=255.0f;
    if (target._g>255.0f) target._g=255.0f;
    if (target._b>255.0f) target._b=255.0f;

    int color = getColor(target);
    binPixels[ii] = color;

    __syncthreads();

    // OK, we wrote the color to the shared memory. Now bring the results to global memory, in a nice coalesced manner.
    int localOfs = 0;
    if (atomicAdd(&flagWriteResults, 1) == 0) {
	// We are the first thread to make it here - do the work.
	unsigned binStartingWriteOffset = 
	    // index of bin (jj) => vertical coordinate of bin (0..RAYCASTBINS-1) => times RAYCASTBINSH => times MAXX
	    ((jj/RAYCASTBINS)*RAYCASTBINSH) * MAXX + 
	    // index of bin (jj) => horizontal coordinate of bin (0..RAYCASTBINS-1) => times RAYCASTBINSW
	    (jj % RAYCASTBINS)*RAYCASTBINSW;
	for(int i=0; i<RAYCASTBINSH; i++) {
	    unsigned writeOffset = binStartingWriteOffset;
	    for(int j=0; j<RAYCASTBINSW; j++)
		pixels[writeOffset++] = binPixels[localOfs++];
	    binStartingWriteOffset += MAXX;
	}
    }
}

// The bridge to the normal C++ world: templated, to include only the mode-specific code in each incantation

template <RenderMode mode>
void CudaRender(
    Matrix3 *cudaWorldToCameraSpace,
    Matrix3 *cudaCameraToLightSpace,
    Matrix3 *cudaWorldToLightSpace,
    Vertex *cudaPtrVertices,
    Triangle *cudaPtrTriangles,
    Vector3 *cudaEyePosInWorldSpace,
    Vector3 *cudaLightPosInWorldSpace,
    Vector3 *cudaLightPosInCameraSpace,
    bool updateLightBins)
{
    int blocksVertices, blocksTriangles;
    blocksVertices  = (g_verticesNo + THREADS_PER_BLOCK - 1) / THREADS_PER_BLOCK;
    blocksTriangles = (g_trianglesNo + THREADS_PER_BLOCK - 1) / THREADS_PER_BLOCK;

    // Triangle bins in camera space
    static int *ptrCudaCameraBoundingBoxesCount = NULL;
    static BoundingBoxCamera *ptrCudaCameraBBoxes = NULL;
    static BoundingBoxCamera *ptrCudaCameraBBoxesPerThreadBlock = NULL;
    static BoundingBoxCamera* ptrHostCameraBBoxesPerBlock = NULL;
    static unsigned *ptrCudaCameraBinsGridCounts = NULL; // RAYCASTBINS * RAYCASTBINS integers, giving the depth of each bin
    static unsigned *ptrCudaCameraBinsGridOffsets = NULL; // RAYCASTBINS * RAYCASTBINS integers with final offset of each bin
    static int *ptrCudaCameraLastBinOffset = NULL;
    static IntersectionCacheCameraspace *ptrCudaCameraIntersectionsCachePerTriangle = NULL;

    // Triangle bins in light space
    static int *ptrCudaLightBoundingBoxesCount = NULL;
    static BoundingBoxLight *ptrCudaLightBBoxes = NULL;
    static BoundingBoxLight *ptrCudaLightBBoxesPerThreadBlock = NULL;
    static BoundingBoxLight* ptrHostLightBBoxesPerBlock = NULL;
    static unsigned *ptrCudaLightBinsGridCounts = NULL; // RAYCASTBINS * RAYCASTBINS integers, giving the depth of each bin
    static unsigned *ptrCudaLightBinsGridOffsets = NULL; // RAYCASTBINS * RAYCASTBINS integers with final offset of each bin
    static int *ptrCudaLightLastBinOffset = NULL;
    static IntersectionCacheLightspace *ptrCudaLightIntersectionsCachePerTriangle = NULL;

    // Allocations done the first time we're called.
    if (!ptrCudaCameraBoundingBoxesCount) {
	SAFE( cudaMalloc((void**)&ptrCudaCameraBoundingBoxesCount, sizeof(int)) );
	SAFE( cudaMalloc((void**)&ptrCudaCameraBBoxes, g_trianglesNo*sizeof(BoundingBoxCamera)) );
	SAFE( cudaMalloc((void**)&ptrCudaCameraBBoxesPerThreadBlock, blocksTriangles*sizeof(BoundingBoxCamera)) );
	ptrHostCameraBBoxesPerBlock = (BoundingBoxCamera*) malloc(sizeof(BoundingBoxCamera)*blocksTriangles);
	SAFE( cudaMalloc((void**)&ptrCudaCameraBinsGridCounts, RAYCASTBINS*RAYCASTBINS*sizeof(unsigned)) );
	SAFE( cudaMalloc((void**)&ptrCudaCameraBinsGridOffsets, RAYCASTBINS*RAYCASTBINS*sizeof(unsigned)) );
	SAFE( cudaMalloc((void**)&ptrCudaCameraLastBinOffset, sizeof(int)) );
	SAFE( cudaMalloc((void**)&ptrCudaCameraIntersectionsCachePerTriangle, g_trianglesNo*sizeof(IntersectionCacheCameraspace)) );

	SAFE( cudaMalloc((void**)&ptrCudaLightBoundingBoxesCount, sizeof(int)) );
	SAFE( cudaMalloc((void**)&ptrCudaLightBBoxes, g_trianglesNo*sizeof(BoundingBoxLight)) );
	SAFE( cudaMalloc((void**)&ptrCudaLightBBoxesPerThreadBlock, blocksTriangles*sizeof(BoundingBoxLight)) );
	ptrHostLightBBoxesPerBlock = (BoundingBoxLight*) malloc(sizeof(BoundingBoxLight)*blocksTriangles);
	SAFE( cudaMalloc((void**)&ptrCudaLightBinsGridCounts, RAYCASTBINS*RAYCASTBINS*sizeof(unsigned)) );
	SAFE( cudaMalloc((void**)&ptrCudaLightBinsGridOffsets, RAYCASTBINS*RAYCASTBINS*sizeof(unsigned)) );
	SAFE( cudaMalloc((void**)&ptrCudaLightLastBinOffset, sizeof(int)) );
	SAFE( cudaMalloc((void**)&ptrCudaLightIntersectionsCachePerTriangle, g_trianglesNo*sizeof(IntersectionCacheLightspace)) );
    }

    int *pixels;
    glBindTexture(GL_TEXTURE_2D, tex);
    SAFE(cudaGLMapBufferObject((void**)&pixels, buffer));

    // Clear all pixels to ambient
    cudaMemset(pixels, 0x40, MAXX*MAXY*sizeof(unsigned));

    if (mode == MODE_POINTS)
	CoreLoopVertices<<< blocksVertices, THREADS_PER_BLOCK >>>(
	    pixels, cudaWorldToCameraSpace, cudaPtrVertices, cudaEyePosInWorldSpace);
    else if (mode == MODE_POINTSHIDDEN)
	CoreLoopTriangles<<< blocksTriangles, THREADS_PER_BLOCK >>>(
	    pixels, cudaWorldToCameraSpace, cudaPtrVertices, cudaPtrTriangles, cudaEyePosInWorldSpace);
    else if (mode == MODE_RAYCASTER || mode == MODE_RAYCASTER_SHADOWS) {
	// First, calculate the triangles' bounding boxes and store them in bins.
	//
	// Unfortunately, in CUDA there's no convenient STL-like containers that auto-grow.
	// We therefore do the work in multiple passes:
	//
	// Pass 1: Just count how many bins we will need, by transforming the triangles
	//         and (atomically) increasing a counter.
	//
	// Count for camera space:
	//
	cudaMemset(ptrCudaCameraBoundingBoxesCount, 0, sizeof(int)); // reset the counter
	int hostCameraBoundingBoxesCount = 0; // host-side of the counter
	DBG_PUTS(1, "CalculateTrianglesCameraBoundingBoxes begins");
	CalculateTrianglesCameraBoundingBoxes<<< blocksTriangles, THREADS_PER_BLOCK >>>(
	    cudaWorldToCameraSpace,
	    cudaPtrVertices, cudaPtrTriangles,
	    cudaEyePosInWorldSpace,
	    ptrCudaCameraBoundingBoxesCount, ptrCudaCameraBBoxes, ptrCudaCameraIntersectionsCachePerTriangle);
	DBG_PUTS(1, "CalculateTrianglesCameraBoundingBoxes ends");
	// read host-side of the counter
	SAFE( cudaMemcpy(
	    &hostCameraBoundingBoxesCount, ptrCudaCameraBoundingBoxesCount, sizeof(int), cudaMemcpyDeviceToHost) );
	DBG_PRINTF(1, "Visible triangles in camera space: %d\n", hostCameraBoundingBoxesCount);

	static int hostLightBoundingBoxesCount = 0;
	if (updateLightBins) {
	    // Count for light space:
	    //
	    cudaMemset(ptrCudaLightBoundingBoxesCount, 0, sizeof(int));
	    DBG_PUTS(1, "CalculateTrianglesLightBoundingBoxes begins");
	    CalculateTrianglesLightBoundingBoxes<<< blocksTriangles, THREADS_PER_BLOCK >>>(
		    cudaWorldToLightSpace,
		    cudaPtrVertices, cudaPtrTriangles,
		    cudaLightPosInWorldSpace,
		    ptrCudaLightBoundingBoxesCount, ptrCudaLightBBoxes, ptrCudaLightIntersectionsCachePerTriangle);
	    DBG_PUTS(1, "CalculateTrianglesLightBoundingBoxes ends");
	    // read host-side of the counter
	    SAFE( cudaMemcpy(
		&hostLightBoundingBoxesCount, ptrCudaLightBoundingBoxesCount, sizeof(int), cudaMemcpyDeviceToHost) );
	    DBG_PRINTF(1, "Visible triangles in light space: %d\n", hostLightBoundingBoxesCount);
	}

	// Pass 2: find the global bounding box, via parallel reduction (ie. many threads doing min/max)
	//
	// Reduction for camera space:
	//
	int blocksCameraBoundingBoxes = (hostCameraBoundingBoxesCount + THREADS_PER_BLOCK - 1)/THREADS_PER_BLOCK;
	DBG_PUTS(1, "ParallelReductionOfBoundingBoxes (camera) begins");
	ParallelReductionOfBoundingBoxes<<< blocksCameraBoundingBoxes, THREADS_PER_BLOCK >>>(
	    ptrCudaCameraBBoxes, ptrCudaCameraBBoxesPerThreadBlock, ptrCudaCameraBoundingBoxesCount);
	DBG_PUTS(1, "ParallelReductionOfBoundingBoxes (camera) ends");
	// ...then do a final reduction at host side:
	SAFE( cudaMemcpy(
	    ptrHostCameraBBoxesPerBlock,
	    ptrCudaCameraBBoxesPerThreadBlock,
	    blocksTriangles*sizeof(BoundingBoxCamera),
	    cudaMemcpyDeviceToHost) );
	for(int i=1; i<blocksCameraBoundingBoxes; i++) {
	    DBG_PRINTF(2, "bbox %d: %f %f %f %f\n",
		i,
		ptrHostCameraBBoxesPerBlock[i]._bx, ptrHostCameraBBoxesPerBlock[i]._by,
		ptrHostCameraBBoxesPerBlock[i]._tx, ptrHostCameraBBoxesPerBlock[i]._ty);
            ptrHostCameraBBoxesPerBlock[0]._bx = fminf(
		ptrHostCameraBBoxesPerBlock[0]._bx, ptrHostCameraBBoxesPerBlock[i]._bx);
            ptrHostCameraBBoxesPerBlock[0]._by = fminf(
		ptrHostCameraBBoxesPerBlock[0]._by, ptrHostCameraBBoxesPerBlock[i]._by);
            ptrHostCameraBBoxesPerBlock[0]._tx = fmaxf(
		ptrHostCameraBBoxesPerBlock[0]._tx, ptrHostCameraBBoxesPerBlock[i]._tx);
            ptrHostCameraBBoxesPerBlock[0]._ty = fmaxf(
		ptrHostCameraBBoxesPerBlock[0]._ty, ptrHostCameraBBoxesPerBlock[i]._ty);
	}
	// ...and calculate the "quantization" factor to use for our target number of bins (RAYCASTBINS)
	coord binSpanCameraX = (ptrHostCameraBBoxesPerBlock[0]._tx - ptrHostCameraBBoxesPerBlock[0]._bx)/RAYCASTBINS;
	coord binSpanCameraY = (ptrHostCameraBBoxesPerBlock[0]._ty - ptrHostCameraBBoxesPerBlock[0]._by)/RAYCASTBINS;
	DBG_PRINTF(1, "camera space bbox: %f %f %f %f\n",
	    ptrHostCameraBBoxesPerBlock[0]._bx, ptrHostCameraBBoxesPerBlock[0]._by,
	    ptrHostCameraBBoxesPerBlock[0]._tx, ptrHostCameraBBoxesPerBlock[0]._ty);
	DBG_PRINTF(1, "camera space binSpans: %f %f\n", binSpanCameraX, binSpanCameraY);

	static int blocksLightBoundingBoxes;
	static coord binSpanLightX;
	static coord binSpanLightY;
	if (updateLightBins) {
	    //
	    // Reduction for light space:
	    //
	    blocksLightBoundingBoxes = (hostLightBoundingBoxesCount + THREADS_PER_BLOCK - 1)/THREADS_PER_BLOCK;
	    DBG_PUTS(1, "ParallelReductionOfBoundingBoxes (light) begins");
	    ParallelReductionOfBoundingBoxes<<< blocksLightBoundingBoxes, THREADS_PER_BLOCK >>>(
		ptrCudaLightBBoxes, ptrCudaLightBBoxesPerThreadBlock, ptrCudaLightBoundingBoxesCount);
	    DBG_PUTS(1, "ParallelReductionOfBoundingBoxes (light) ends");
	    // ...then do a final reduction at host side:
	    SAFE( cudaMemcpy(
		ptrHostLightBBoxesPerBlock,
		ptrCudaLightBBoxesPerThreadBlock,
		blocksTriangles*sizeof(BoundingBoxLight),
		cudaMemcpyDeviceToHost) );
	    for(int i=1; i<blocksLightBoundingBoxes; i++) {
		ptrHostLightBBoxesPerBlock[0]._bx = fminf(
		    ptrHostLightBBoxesPerBlock[0]._bx, ptrHostLightBBoxesPerBlock[i]._bx);
		ptrHostLightBBoxesPerBlock[0]._by = fminf(
		    ptrHostLightBBoxesPerBlock[0]._by, ptrHostLightBBoxesPerBlock[i]._by);
		ptrHostLightBBoxesPerBlock[0]._tx = fmaxf(
		    ptrHostLightBBoxesPerBlock[0]._tx, ptrHostLightBBoxesPerBlock[i]._tx);
		ptrHostLightBBoxesPerBlock[0]._ty = fmaxf(
		    ptrHostLightBBoxesPerBlock[0]._ty, ptrHostLightBBoxesPerBlock[i]._ty);
	    }
	    // ...and calculate the "quantization" factor to use for our target number of bins (RAYCASTBINS)
	    binSpanLightX = (ptrHostLightBBoxesPerBlock[0]._tx - ptrHostLightBBoxesPerBlock[0]._bx)/RAYCASTBINS;
	    binSpanLightY = (ptrHostLightBBoxesPerBlock[0]._ty - ptrHostLightBBoxesPerBlock[0]._by)/RAYCASTBINS;
	    DBG_PRINTF(1, "light space bbox: %f %f %f %f\n",
		ptrHostLightBBoxesPerBlock[0]._bx, ptrHostLightBBoxesPerBlock[0]._by,
		ptrHostLightBBoxesPerBlock[0]._tx, ptrHostLightBBoxesPerBlock[0]._ty);
	    DBG_PRINTF(1, "light space binSpans: %f %f\n", binSpanLightX, binSpanLightY);
	}

	// Pass 3:  Now that we've calculated the bin quantization factor, we measure how many
	//          triangles will fall into each bin.
	//
	// In camera space:
	//
	// 1. Reset the counter for each bin
	cudaMemset(ptrCudaCameraBinsGridCounts, 0, RAYCASTBINS*RAYCASTBINS*sizeof(unsigned));
	// 2. Count the number of triangles to place in each bin
	DBG_PUTS(1, "CountBinTriangleListDepths (camera) begins");
	CountBinTriangleListDepths<<< blocksCameraBoundingBoxes, THREADS_PER_BLOCK >>>(
	    binSpanCameraX, binSpanCameraY, ptrHostCameraBBoxesPerBlock[0]._bx, ptrHostCameraBBoxesPerBlock[0]._by,
	    ptrCudaCameraBBoxes, ptrCudaCameraBinsGridCounts, ptrCudaCameraBoundingBoxesCount);
	DBG_PUTS(1, "CountBinTriangleListDepths (camera) ends");

	if (updateLightBins) {
	    // In light space:
	    //
	    // 1. Reset the counter for each bin
	    cudaMemset(ptrCudaLightBinsGridCounts, 0, RAYCASTBINS*RAYCASTBINS*sizeof(unsigned));
	    // 2. Count the number of triangles to place in each bin
	    DBG_PUTS(1, "CountBinTriangleListDepths (light) begins");
	    CountBinTriangleListDepths<<< blocksLightBoundingBoxes, THREADS_PER_BLOCK >>>(
		binSpanLightX, binSpanLightY, ptrHostLightBBoxesPerBlock[0]._bx, ptrHostLightBBoxesPerBlock[0]._by,
		ptrCudaLightBBoxes, ptrCudaLightBinsGridCounts, ptrCudaLightBoundingBoxesCount);
	    DBG_PUTS(1, "CountBinTriangleListDepths (light) ends");
	}

	// Pass 4: We now accumulate the size of each bin to all the previous ones, to form
	// the bin offsets in the "heap"

	// Both light and camera have RAYCASTBINS x RAYCASTBINS:
	int blockBins = (RAYCASTBINS*RAYCASTBINS + THREADS_PER_BLOCK - 1)/THREADS_PER_BLOCK;

	//
	// In camera space:
	//
	DBG_PUTS(1, "CalculateBinOffsets (camera) begins");
	CalculateBinOffsets<<< blockBins, THREADS_PER_BLOCK>>>(
	    ptrCudaCameraBinsGridCounts, ptrCudaCameraBinsGridOffsets, ptrCudaCameraLastBinOffset);
	DBG_PUTS(1, "CalculateBinOffsets (camera) ends");
	int totalCameraBinsRequired = 0;
	SAFE( cudaMemcpy(
	    &totalCameraBinsRequired,
	    ptrCudaCameraLastBinOffset,
	    sizeof(int),
	    cudaMemcpyDeviceToHost) );
	DBG_PRINTF(1, "Total camera slots required: %d\n", totalCameraBinsRequired);
	int *ptrCudaCameraBins = NULL;
	SAFE( cudaMalloc((void**)&ptrCudaCameraBins, totalCameraBinsRequired*sizeof(int)) );

	static int *ptrCudaLightBins = NULL;
	if (updateLightBins) {
	    //
	    // In light space:
	    //
	    DBG_PUTS(1, "CalculateBinOffsets (light) begins");
	    CalculateBinOffsets<<< blockBins, THREADS_PER_BLOCK>>>(
		ptrCudaLightBinsGridCounts, ptrCudaLightBinsGridOffsets, ptrCudaLightLastBinOffset);
	    DBG_PUTS(1, "CalculateBinOffsets (light) ends");
	    int totalLightBinsRequired = 0;
	    SAFE( cudaMemcpy(
		&totalLightBinsRequired,
		ptrCudaLightLastBinOffset,
		sizeof(int),
		cudaMemcpyDeviceToHost) );
	    DBG_PRINTF(1, "Total light slots required: %d\n", totalLightBinsRequired);
	    if (ptrCudaLightBins) {
		SAFE(cudaFree(ptrCudaLightBins));
	    }
	    SAFE( cudaMalloc((void**)&ptrCudaLightBins, totalLightBinsRequired*sizeof(int)) );
	}

	// Pass 5: And finally, we can now place the triangle indexes in the bin lists.
	//
	// In camera space:
	//
	// "ptrCudaCameraBins" point to totalCameraBinsRequired Triangle indexes, i.e. the "heap" with the lists
	// of candidate triangles
	DBG_PUTS(1, "PlaceTrianglesInBins (camera) begins");
	PlaceTrianglesInBins<<< blocksCameraBoundingBoxes, THREADS_PER_BLOCK>>>(
	    binSpanCameraX, binSpanCameraY, ptrHostCameraBBoxesPerBlock[0]._bx, ptrHostCameraBBoxesPerBlock[0]._by,
	    ptrCudaCameraBBoxes, ptrCudaCameraBinsGridOffsets, ptrCudaCameraBins, ptrCudaCameraBoundingBoxesCount);
	DBG_PUTS(1, "PlaceTrianglesInBins (camera) ends");

	if (updateLightBins) {
	    // In light space:
	    //
	    // "ptrCudaLightBins" point to totalLightBinsRequired Triangle indexes, i.e. the "heap" with the lists
	    // of candidate triangles
	    DBG_PUTS(1, "PlaceTrianglesInBins (light) begins");
	    PlaceTrianglesInBins<<< blocksLightBoundingBoxes, THREADS_PER_BLOCK>>>(
		binSpanLightX, binSpanLightY, ptrHostLightBBoxesPerBlock[0]._bx, ptrHostLightBBoxesPerBlock[0]._by,
		ptrCudaLightBBoxes, ptrCudaLightBinsGridOffsets, ptrCudaLightBins, ptrCudaLightBoundingBoxesCount);
	    DBG_PUTS(1, "PlaceTrianglesInBins (light) ends");
	}

	// Finally, use the bins to accelerate a ray casting of the screen
	int blockPixels = (MAXY*MAXX + RAYCASTBINPIXELS - 1)/RAYCASTBINPIXELS;
	DBG_PUTS(1, "CoreLoopTrianglesRaycaster begins");
	if (mode == MODE_RAYCASTER)
	    CoreLoopTrianglesRaycaster<false><<< blockPixels, RAYCASTBINPIXELS >>>(
		pixels,
		cudaWorldToCameraSpace, cudaCameraToLightSpace,
		cudaPtrVertices, cudaPtrTriangles,
		cudaEyePosInWorldSpace, cudaLightPosInWorldSpace, cudaLightPosInCameraSpace,
		binSpanCameraX, binSpanCameraY, ptrHostCameraBBoxesPerBlock[0]._bx, ptrHostCameraBBoxesPerBlock[0]._by,
		binSpanLightX, binSpanLightY, ptrHostLightBBoxesPerBlock[0]._bx, ptrHostLightBBoxesPerBlock[0]._by,
		ptrCudaCameraBins, ptrCudaCameraBinsGridOffsets, ptrCudaCameraBoundingBoxesCount,
		ptrCudaLightBins, ptrCudaLightBinsGridOffsets, ptrCudaLightBoundingBoxesCount,
		ptrCudaCameraIntersectionsCachePerTriangle, ptrCudaCameraBBoxes,
		ptrCudaLightIntersectionsCachePerTriangle, ptrCudaLightBBoxes,
		ptrCudaCameraLastBinOffset, ptrCudaLightLastBinOffset);
	else // (mode == MODE_RAYCASTER_SHADOWS)
	    CoreLoopTrianglesRaycaster<true><<< blockPixels, RAYCASTBINPIXELS >>>(
		pixels,
		cudaWorldToCameraSpace, cudaCameraToLightSpace,
		cudaPtrVertices, cudaPtrTriangles,
		cudaEyePosInWorldSpace, cudaLightPosInWorldSpace, cudaLightPosInCameraSpace,
		binSpanCameraX, binSpanCameraY, ptrHostCameraBBoxesPerBlock[0]._bx, ptrHostCameraBBoxesPerBlock[0]._by,
		binSpanLightX, binSpanLightY, ptrHostLightBBoxesPerBlock[0]._bx, ptrHostLightBBoxesPerBlock[0]._by,
		ptrCudaCameraBins, ptrCudaCameraBinsGridOffsets, ptrCudaCameraBoundingBoxesCount,
		ptrCudaLightBins, ptrCudaLightBinsGridOffsets, ptrCudaLightBoundingBoxesCount,
		ptrCudaCameraIntersectionsCachePerTriangle, ptrCudaCameraBBoxes,
		ptrCudaLightIntersectionsCachePerTriangle, ptrCudaLightBBoxes,
		ptrCudaCameraLastBinOffset, ptrCudaLightLastBinOffset);
	DBG_PUTS(1, "CoreLoopTrianglesRaycaster ends");

	SAFE(cudaFree(ptrCudaCameraBins));
	// the light-space bins are freed and recalculated only when the light position changes, to optimize speed.
	// (updateLightBins is set to true by the main() when that happens)
    }

    SAFE(cudaThreadSynchronize());
    SAFE(cudaGLUnmapBufferObject(buffer));

    // Use OpenGL texture to draw this at lightning speed (the PBO buffer is already on the card, no useless bus criss-cross)
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, buffer);
    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, MAXX, MAXY, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glBindBuffer(GL_PIXEL_UNPACK_BUFFER, 0);

    glEnable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f, 1.0f); glVertex2f(-1.0f, -1.0f);
    glTexCoord2f(1.0f, 1.0f); glVertex2f(1.0f, -1.0f);
    glTexCoord2f(1.0f, 0.0f); glVertex2f(1.0f, 1.0f);
    glTexCoord2f(0.0f, 0.0f); glVertex2f(-1.0f, 1.0f);
    glEnd();

    glDisable(GL_LIGHTING);
    glDisable(GL_TEXTURE_2D);
    glColor3f(1.f, 1.f, 1.f);
    glRasterPos2f(-0.95, 0.9);
    const char *help = "Press H for help";
    for(int o=0;o<strlen(help); o++)
	glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, help[o]);

    SDL_GL_SwapBuffers();
}

template void CudaRender<MODE_POINTS>(
    Matrix3 *cudaWorldToCameraSpace, Matrix3 *cudaCameraToLightSpace, Matrix3 *cudaWorldToLightSpace,
    Vertex *cudaPtrVertices, Triangle *cudaPtrTriangles,
    Vector3 *cudaEyePosInWorldSpace, Vector3 *pLightInWorldSpace, Vector3 *pLightInCameraSpace, bool updateLightBins);
template void CudaRender<MODE_POINTSHIDDEN>(
    Matrix3 *cudaWorldToCameraSpace, Matrix3 *cudaCameraToLightSpace, Matrix3 *cudaWorldToLightSpace,
    Vertex *cudaPtrVertices, Triangle *cudaPtrTriangles,
    Vector3 *cudaEyePosInWorldSpace, Vector3 *pLightInWorldSpace, Vector3 *pLightInCameraSpace, bool updateLightBins);
template void CudaRender<MODE_RAYCASTER>(
    Matrix3 *cudaWorldToCameraSpace, Matrix3 *cudaCameraToLightSpace, Matrix3 *cudaWorldToLightSpace,
    Vertex *cudaPtrVertices, Triangle *cudaPtrTriangles,
    Vector3 *cudaEyePosInWorldSpace, Vector3 *pLightInWorldSpace, Vector3 *pLightInCameraSpace, bool updateLightBins);
template void CudaRender<MODE_RAYCASTER_SHADOWS>(
    Matrix3 *cudaWorldToCameraSpace, Matrix3 *cudaCameraToLightSpace, Matrix3 *cudaWorldToLightSpace,
    Vertex *cudaPtrVertices, Triangle *cudaPtrTriangles,
    Vector3 *cudaEyePosInWorldSpace, Vector3 *pLightInWorldSpace, Vector3 *pLightInCameraSpace, bool updateLightBins);
