/*
 *  renderer - A simple implementation of polygon-based 3D algorithms.
 *  Copyright (C) 2004  Thanassis Tsiodras (ttsiodras@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __CUDA_RENDERER_H__
#define __CUDA_RENDERER_H__

#include <config.h>

#include "Types.h"
#include "Keyboard.h"
#include "Base3d.h"
#include "Camera.h" 

// Constants

// I cannot use C++ constants for these, since I am using them to allocate
// stack arrays or shared memory in CUDA kernels.

#define THREADS_PER_BLOCK   128

#define MAXX		    768
#define MAXY		    512
#define RAYCASTBINS	    128
#define RAYCASTBINSW	    (MAXX/RAYCASTBINS)
#define RAYCASTBINSH	    (MAXY/RAYCASTBINS)
#define RAYCASTBINPIXELS    ((RAYCASTBINSH)*(RAYCASTBINSW))

#define AMBIENT		96.f
#define DIFFUSE		128.f
#define SPECULAR	192.f

#define FOV		1024

// Debugging helpers

#define DEBUG_LEVEL 0

#define DBG_PRINTF(level, fmt, ...) \
    do { if (level <= DEBUG_LEVEL) { printf(fmt, __VA_ARGS__); fflush(stdout); }} while (0)

#define DBG_PUTS(level, msg) \
    do { if (level <= DEBUG_LEVEL) { puts(msg); fflush(stdout); }} while (0)

#  define SAFE_NO_SYNC( call ) {                                     \
    cudaError_t err = call;                                                  \
    if( cudaSuccess != err) {                                               \
        fprintf(stderr, "Cuda driver error %x in file '%s' in line %i.\n",   \
                err, __FILE__, __LINE__ );                                   \
        exit(EXIT_FAILURE);                                                  \
    } }

#  define SAFE( call )       SAFE_NO_SYNC(call);

// Macro for global variables

#ifdef DEFINE_GLOBALS
#define DECLARE_DEFINE_GLOBAL(t,var,val) t var = val;
#else
#define DECLARE_DEFINE_GLOBAL(t,var,val) extern t var;
#endif

// Enumerations

// The rendering mode we are in
enum RenderMode {
    MODE_POINTS = 0,
    MODE_POINTSHIDDEN = 1,
    MODE_RAYCASTER = 2,
    MODE_RAYCASTER_SHADOWS = 3
};

// Global variables

DECLARE_DEFINE_GLOBAL(unsigned, g_verticesNo, 0)
DECLARE_DEFINE_GLOBAL(Vertex *, g_vertices, NULL)
DECLARE_DEFINE_GLOBAL(unsigned, g_trianglesNo, 0)
DECLARE_DEFINE_GLOBAL(Triangle *, g_triangles, NULL)

DECLARE_DEFINE_GLOBAL(GLuint,buffer,0);
DECLARE_DEFINE_GLOBAL(GLuint,tex,0);

// The gateway to CUDA, called from C++ (src/main.cpp)
template <enum RenderMode>
void CudaRender(
    Matrix3 *pMv, Matrix3 *cudaCameraToLightSpace, Matrix3 *cudaWorldToLightSpace,
    Vertex *cudaPtrVertices, Triangle *cudaPtrTriangles,
    Vector3 *eye, Vector3 *pLightInWorldSpace, Vector3 *pLightInCameraSpace, bool updateLightBins);

#endif
