/*
 *  renderer - A simple implementation of polygon-based 3D algorithms.
 *  Copyright (C) 2004  Thanassis Tsiodras (ttsiodras@gmail.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <config.h>

#ifdef _WIN32
#include <windows.h>
#include <GL/glew.h>
#else
#define GL_GLEXT_PROTOTYPES
#include <GL/glew.h>
#include <GL/gl.h>
#endif
#include <GL/glut.h>

#include <iostream>

#include <math.h>

#define GL_GLEXT_PROTOTYPES
#include <GL/gl.h>

#include <SDL.h>

#include <cuda.h>
#include <cuda_runtime.h>
#include <cuda_gl_interop.h>

#define DEFINE_GLOBALS
#include "cudarenderer.h"

#include "Clock.h"
#include "Loader.h"
#include "Utility.h"

#ifndef M_PI
#define M_PI 3.14156265
#endif

using namespace std;

// The CUDA kernels will draw directly inside an OpenGL buffer (PBO)
// which will be displayed via an OpenGL texture.

// OpenGL buffer and texture helper functions

void create_buffer(GLuint* buffer)
{
    glGenBuffersARB(1, buffer);
    glBindBuffer(GL_PIXEL_PACK_BUFFER, *buffer);
    glBufferData(GL_PIXEL_PACK_BUFFER, MAXX*MAXY*4, NULL, GL_DYNAMIC_DRAW);
    glBindBuffer(GL_PIXEL_PACK_BUFFER, 0);
}

void destroy_buffer(GLuint* buffer)
{
    glBindBuffer(GL_TEXTURE_2D, 0);
    glDeleteBuffers(1, buffer);
    *buffer = 0;
}

void create_texture(GLuint* tex)
{
    glGenTextures(1, tex);
    glBindTexture(GL_TEXTURE_2D, *tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, MAXX, MAXY, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
}

void destroy_texture(GLuint* tex)
{
    glBindTexture(GL_TEXTURE_2D, *tex);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 0, 0, 0, GL_RGBA, GL_UNSIGNED_BYTE, NULL);
    glBindTexture(GL_TEXTURE_2D, 0);
    glDeleteTextures(1, tex);
}

int main(int argc, char *argv[])
{
    bool bench = false;
    int benchFrames = 0;

    // GLUT may remove from args
    glutInit(&argc,argv);

    if (argc < 2)
	panic("Usage: %s [-b frames] <filename.tri>\n", argv[0]);

    // In case we want to benchmark
    if (!strcmp(argv[1], "-b")) {
	bench = true;
	benchFrames = atoi(argv[2]);
    }

    float maxi = load_object(argv[argc-1]);

    // These checks should be compile-time asserts, but I am lazy
    if (MAXX != RAYCASTBINSW*RAYCASTBINS)
	panic("MAXX must be an exact multiple of RAYCASTBINS");
    if (MAXY != RAYCASTBINSH*RAYCASTBINS)
	panic("MAXY must be an exact multiple of RAYCASTBINS");

    // SDL startup

    if ( SDL_Init(SDL_INIT_VIDEO) < 0 )
        panic("Couldn't initialize SDL: %s\n", SDL_GetError());
    atexit(SDL_Quit); // Clean up on exit
    if (!SDL_SetVideoMode( MAXX, MAXY, 0, SDL_OPENGL))
        panic("Couldn't set video mode: %s\n", SDL_GetError());
    // Ignore mice, we love keyboards
    SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);

    // OpenGL startup

    // We need an OpenGL PBO...
    cout << "Loading extensions: " << glewGetErrorString(glewInit()) << endl;
    if (!glewIsSupported( "GL_VERSION_1_5 GL_ARB_pixel_buffer_object" ))
	panic("Missing GL_ARB_pixel_buffer_object\n");

    glViewport(0, 0, MAXX, MAXY);
    glClearColor(0.3f, 0.3f, 0.3f, 0.5f);
    glEnable(GL_TEXTURE_2D);
    glLoadIdentity();

    create_buffer(&buffer);
    create_texture(&tex);
    SAFE( cudaGLRegisterBufferObject(buffer) ); // Tell CUDA about the PBO

    // Now, allocate the CUDA side of the data (in CUDA global memory)

    Vertex *cudaVertices;
    Triangle *cudaTriangles;

    SAFE( cudaMalloc((void**)&cudaVertices, g_verticesNo*sizeof(Vertex)) );
    SAFE( cudaMalloc((void**)&cudaTriangles, g_trianglesNo*sizeof(Triangle)) );
    SAFE( cudaMemcpy(cudaVertices, g_vertices, g_verticesNo*sizeof(Vertex), cudaMemcpyHostToDevice) );
    SAFE( cudaMemcpy(cudaTriangles, g_triangles, g_trianglesNo*sizeof(Triangle), cudaMemcpyHostToDevice) );
    SAFE( cudaMemcpyToSymbol("VERTICES", &g_verticesNo, sizeof(int)) );
    SAFE( cudaMemcpyToSymbol("TRIANGLES", &g_trianglesNo, sizeof(int)) );

    float angle1=0.0f;
    float angle2=0.0f;
    float angle3=45.0f*M_PI/180.f;
    const coord EyeDistanceFactor = 4.0;
    const coord LightDistanceFactor = 4.0;

    Vector3 light(LightDistanceFactor*maxi, LightDistanceFactor*maxi, LightDistanceFactor*maxi);
    Vector3 *pLight = &light;
    Vector3 lightInCameraSpace; // x-formed each frame 

    int framesDrawn = 0;
    bool autoRotate = true;

    Keyboard keys;
    Point eye(maxi*EyeDistanceFactor, 0.0, 0.0);
    Point lookat(eye._x + 1.0f*cos(angle2)*cos(angle1),
		 eye._y + 1.0f*cos(angle2)*sin(angle1),
		 eye._z + 1.0f*sin(angle2));

    Camera sony(eye, lookat);

    Clock watch;

#define DEGREES_TO_RADIANS(x) ((coord)((x)*M_PI/180.0))
    coord dAngle = DEGREES_TO_RADIANS(0.3f);

    keys.poll();

    // Allocate more CUDA-side data (global memory) for positions and transformations matrices
    Matrix3 *cudaSony = NULL;
    SAFE( cudaMalloc((void**)&cudaSony, sizeof(sony._mv)) );
    Matrix3 *cudaCameraToLightSpace = NULL;
    SAFE( cudaMalloc((void**)&cudaCameraToLightSpace, sizeof(Matrix3)) );
    Matrix3 *cudaWorldToLightSpace = NULL;
    SAFE( cudaMalloc((void**)&cudaWorldToLightSpace, sizeof(Matrix3)) );
    Vector3 *cudaEye = NULL;
    SAFE( cudaMalloc((void**)&cudaEye, sizeof(Vector3)) );
    Vector3 *cudaLightInCameraSpace = NULL;
    Vector3 *cudaLightInWorldSpace = NULL;
    SAFE( cudaMalloc((void**)&cudaLightInCameraSpace, sizeof(Vector3)) );
    SAFE( cudaMalloc((void**)&cudaLightInWorldSpace, sizeof(Vector3)) );

    int mode = MODE_RAYCASTER_SHADOWS;
    bool lightMoved = true; // Force recalculation of light bins (a) the first time,
			    // and (b) when light is moved

    pLight->_x = LightDistanceFactor*maxi*cos(angle3);
    pLight->_y = LightDistanceFactor*maxi*sin(angle3);

    SDL_WM_SetCaption("Real-time raytracing", "Real-time raytracing");
    watch.reset();
    while(!keys.Abort()) {
	framesDrawn++;
	if (bench && framesDrawn>benchFrames)
	    break;
	if (!bench) {
	    // Only allow keyboard control if we are not benchmarking
	    if (keys.Left())
		angle1-=dAngle;
	    if (keys.Right())
		angle1+=dAngle;
	    if (keys.Up())
		angle2=min(angle2+dAngle, DEGREES_TO_RADIANS(89.0f));
	    if (keys.Down())
		angle2=max(angle2-dAngle, DEGREES_TO_RADIANS(-89.0f));
	    if (keys.Forward() || keys.Backward()) {
		Vector3 fromEyeToLookat(lookat);
		fromEyeToLookat -= eye;
		if (autoRotate)
		    fromEyeToLookat *= 0.05f;
		else
		    fromEyeToLookat *= 0.05f*maxi;
		/*
		cout << "At " << eye._x << " " << eye._y << " " << eye._z << endl;
		cout << "Looking at " << lookat._x << " " << lookat._y << " " << lookat._z << endl;
		cout << "Adding " << fromEyeToLookat._x << " " << fromEyeToLookat._y << " " << fromEyeToLookat._z << endl;
		*/
		if (keys.Forward())
		    eye += fromEyeToLookat;
		else
		    eye -= fromEyeToLookat;
	    }
	    if (keys.isS() || keys.isF() || keys.isE() || keys.isD()) {
		Vector3 eyeToLookatPoint = lookat;
		eyeToLookatPoint -= eye;
		eyeToLookatPoint.normalize();
		Vector3 zenith(0., 0., 1.);
		Vector3 rightAxis = cross(eyeToLookatPoint, zenith);
		rightAxis.normalize();
		Vector3 upAxis = cross(rightAxis, eyeToLookatPoint);
		upAxis.normalize();
		if (keys.isS()) { rightAxis *= 0.05f*maxi; eye -= rightAxis; }
		if (keys.isF()) { rightAxis *= 0.05f*maxi; eye += rightAxis; }
		if (keys.isD()) { upAxis *= 0.05f*maxi; eye -= upAxis; }
		if (keys.isE()) { upAxis *= 0.05f*maxi; eye += upAxis; }
	    }
	    if (keys.isR()) {
		while(keys.isR())
		    keys.poll();
		autoRotate = !autoRotate;
		if (!autoRotate) {
		    Vector3 eyeToAxes = eye;
		    eyeToAxes.normalize();
#if defined(__GNUC__)
		    // Hideous GCC code generation bug - unless I print them, eyeToAxes are not normalized!
		    cout << "Moving to " << eyeToAxes._x << "," <<
			eyeToAxes._y << "," << eyeToAxes._z << endl;
#endif
		    angle2 = asin(-eyeToAxes._z);
		    angle1 = (eye._y<0)?acos(eyeToAxes._x/cos(angle2)):-acos(eyeToAxes._x/cos(angle2));
		} else {
		    angle1 = -angle1;
		    angle2 = -angle2;
		}
		watch.reset();
		framesDrawn = 1;
	    }
	    if (keys.Light()) {
		angle3 += 4*dAngle;
		pLight->_x = LightDistanceFactor*maxi*cos(angle3);
		pLight->_y = LightDistanceFactor*maxi*sin(angle3);
		lightMoved = true;
	    }
	    if (keys.Light2()) {
		angle3 -= 4*dAngle;
		pLight->_x = LightDistanceFactor*maxi*cos(angle3);
		pLight->_y = LightDistanceFactor*maxi*sin(angle3);
		lightMoved = true;
	    }
	    if (keys.isPgDown()) {
		while(keys.isPgDown())
		    keys.poll();
		mode = (mode+3) % 4;
		framesDrawn = 1;
		if (mode == MODE_RAYCASTER_SHADOWS)
		    lightMoved = true;
		watch.reset();
	    }
	    if (keys.isPgUp()) {
		while(keys.isPgUp())
		    keys.poll();
		mode = (mode+1) % 4;
		framesDrawn = 1;
		if (mode == MODE_RAYCASTER_SHADOWS)
		    lightMoved = true;
		watch.reset();
	    }
	    if (keys.isH()) {
		while(keys.isH()) keys.poll();
		glClear(GL_COLOR_BUFFER_BIT);
		glDisable(GL_LIGHTING);
		glDisable(GL_TEXTURE_2D);
		glColor3f(1.f, 1.f, 1.f);
		const char *help =
"KEYS: (available only when not benchmarking)\n"
"\n"
"- Hit 'R' to stop/start auto-rotation of the camera around the object.\n"
"- Fly using the cursor keys,A,Z - and rotate the light with W and Q.\n"
"- S and F are 'strafe' left/right\n"
"- E and D are 'strafe' up/down\n"
"  (strafe keys don't work in auto-rotation mode).\n"
"- PgUp/PgDown change the rendering mode:\n"
"    Points ->\n"
"    CulledTrianglePoints ->\n"
"    Phong raycasting ->\n"
"    Phong and shadows raycasting\n"
"- ESC quits.\n\n"
"Now press H to get out of this help screen\n";

		glRasterPos2f(-0.95, 0.9);
		int lines = 0;
		for(const char *p=help; *p; p++) {
		    if (*p == '\n')
			// 12 point helvetica => 12+3=15 points per line
			glRasterPos2f(-0.95, 0.9 - ((12+3)*2.f/MAXY)*lines++);
		    else
			glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *p);
		}
		SDL_GL_SwapBuffers();
		while(!keys.isH() && !keys.Abort()) keys.poll();
		while(keys.isH() || keys.Abort()) keys.poll();
		framesDrawn = 1;
		watch.reset();
	    }
	}
	if (!autoRotate) {
	    lookat._x = eye._x - 1.0f*cos(angle2)*cos(angle1);
	    lookat._y = eye._y + 1.0f*cos(angle2)*sin(angle1);
	    lookat._z = eye._z + 1.0f*sin(angle2);
	} else {
	    angle1-=dAngle;
	    lookat._x = 0;
	    lookat._y = 0;
	    lookat._z = 0;
	    coord distance = sqrt(eye._x*eye._x + eye._y*eye._y + eye._z*eye._z);
	    eye._x = distance*cos(angle2)*cos(angle1);
	    eye._y = distance*cos(angle2)*sin(angle1);
	    eye._z = distance*sin(angle2);
	}

	// x-form light in cameraspace
	sony.set(eye, lookat);
	Vector3 cameraToLight = *pLight;
	cameraToLight -= eye;
	lightInCameraSpace = sony._mv.multiplyRightWith(cameraToLight);

	// update matrix for x-from from world to lightspace
	Vector3 lightToWorldCenter = *pLight;
	lightToWorldCenter *= -1.f;
	lightToWorldCenter.normalize();
	Vector3 zenith(0., 0., 1.);
	Vector3 rightAxis = cross(lightToWorldCenter, zenith);
	rightAxis.normalize();
	Vector3 upAxis = cross(rightAxis, lightToWorldCenter);
	upAxis.normalize();

	// OK, we calculated the 3 light axis in worldspace.
	// We need them in:
	//
	// camera space:
	Matrix3 cameraToLightSpace;
	cameraToLightSpace._row1 = sony._mv.multiplyRightWith(upAxis);
	cameraToLightSpace._row2 = sony._mv.multiplyRightWith(rightAxis);
	cameraToLightSpace._row3 = sony._mv.multiplyRightWith(lightToWorldCenter);

	// light space:
	Matrix3 worldToLightSpace;
	worldToLightSpace._row1 = upAxis;		// x-axis
	worldToLightSpace._row2 = rightAxis;		// y-axis
	worldToLightSpace._row3 = lightToWorldCenter;	// z-axis

	// Update the CUDA side of the eye/light positions and x-form matrices
	SAFE( cudaMemcpy(cudaSony, &sony._mv, sizeof(sony._mv), cudaMemcpyHostToDevice) );
	SAFE( cudaMemcpy(cudaCameraToLightSpace, &cameraToLightSpace, sizeof(cameraToLightSpace), cudaMemcpyHostToDevice) );
	SAFE( cudaMemcpy(cudaWorldToLightSpace, &worldToLightSpace, sizeof(worldToLightSpace), cudaMemcpyHostToDevice) );
	SAFE( cudaMemcpy(cudaEye, (Vector3*)&sony, sizeof(Vector3), cudaMemcpyHostToDevice) );
	SAFE( cudaMemcpy(cudaLightInWorldSpace, pLight, sizeof(Vector3), cudaMemcpyHostToDevice) );
	SAFE( cudaMemcpy(cudaLightInCameraSpace, &lightInCameraSpace, sizeof(Vector3), cudaMemcpyHostToDevice) );

	// ...and render!
	// C++ templates don't allow run-time variables to be used in template params.
	// In principle, for enumerations (like mode) they could, but who am I to judge.
	if (mode == MODE_POINTS)
	    CudaRender<MODE_POINTS>(
		cudaSony, cudaCameraToLightSpace, cudaWorldToLightSpace,
		cudaVertices, cudaTriangles,
		cudaEye, cudaLightInWorldSpace, cudaLightInCameraSpace, lightMoved);
	else if (mode == MODE_POINTSHIDDEN)
	    CudaRender<MODE_POINTSHIDDEN>(
		cudaSony, cudaCameraToLightSpace, cudaWorldToLightSpace,
		cudaVertices, cudaTriangles,
		cudaEye, cudaLightInWorldSpace, cudaLightInCameraSpace, lightMoved);
	else if (mode == MODE_RAYCASTER)
	    CudaRender<MODE_RAYCASTER>(
		cudaSony, cudaCameraToLightSpace, cudaWorldToLightSpace,
		cudaVertices, cudaTriangles,
		cudaEye, cudaLightInWorldSpace, cudaLightInCameraSpace, lightMoved);
	else if (mode == MODE_RAYCASTER_SHADOWS)
	    CudaRender<MODE_RAYCASTER_SHADOWS>(
		cudaSony, cudaCameraToLightSpace, cudaWorldToLightSpace,
		cudaVertices, cudaTriangles,
		cudaEye, cudaLightInWorldSpace, cudaLightInCameraSpace, lightMoved);

	lightMoved = false;

	keys.poll();

	// The more frames per sec, the smaller the dAngle should be
	// The less frames per sec, the larger  the dAngle should be
	//   so...
	// Move dAngle towards 36/framesPerSec in 15 frames (steps)
	//   for 100fps, dAngle = 0.36
	//   for 10 fps, dAngle = 3.6
	// In both cases, we rotate 360 degrees in 10 seconds
	//dAngle += ((36.0/(framesDrawn/(watch.readMS()/1000.0)))-dAngle)/15.;
	//
	// Changed to 18/framesPerSec (20 seconds needed for a 360 rotation)
	dAngle += (DEGREES_TO_RADIANS(9.0f/(framesDrawn/(watch.readMS()/1000.0f)))-dAngle)/15.0f;
    }
    framesDrawn--;

#ifdef _WIN32
    stringstream speed;
    speed << "Rendering " << framesDrawn << " frames in " << (watch.readMS()/1000.0) << " seconds. (" << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
    MessageBoxA(0, (LPCSTR) speed.str().c_str(), (const char *)"Speed of rendering", MB_OK);
#else
    cout << "Rendering " << framesDrawn << " frames in ";
    cout << (watch.readMS()/1000.0) << " seconds. (";
    cout << framesDrawn/(watch.readMS()/1000.0) << " fps)\n";
#endif

    SAFE(cudaFree(cudaLightInCameraSpace));
    SAFE(cudaFree(cudaLightInWorldSpace));
    SAFE(cudaFree(cudaEye));
    SAFE(cudaFree(cudaSony));
    SAFE(cudaFree(cudaVertices));
    SAFE(cudaFree(cudaTriangles));
    SAFE(cudaGLUnregisterBufferObject(buffer));
    destroy_texture(&tex);
    destroy_buffer(&buffer);

    return 0;
}
